package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.core.compiler.CharOperation;

public class CaptureBinding18 extends CaptureBinding
{
    TypeBinding[] upperBounds;
    private char[] originalName;
    private CaptureBinding18 prototype;
    int recursionLevel;
    
    public CaptureBinding18(ReferenceBinding contextType, char[] sourceName, char[] originalName, int start, int end, int captureID, LookupEnvironment environment) {
        super(contextType, sourceName, start, end, captureID, environment);
        this.recursionLevel = 0;
        this.originalName = originalName;
        this.prototype = this;
    }
    
    private CaptureBinding18(CaptureBinding18 prototype) {
        super(prototype);
        this.recursionLevel = 0;
        this.sourceName = CharOperation.append(prototype.sourceName, '\'');
        this.originalName = prototype.originalName;
        this.upperBounds = prototype.upperBounds;
        this.prototype = prototype.prototype;
    }
    
    public boolean setUpperBounds(TypeBinding[] upperBounds, ReferenceBinding javaLangObject) {
        this.upperBounds = upperBounds;
        if (upperBounds.length > 0) {
            this.firstBound = upperBounds[0];
        }
        int numReferenceInterfaces = 0;
        if (!ReferenceBinding.isConsistentIntersection(upperBounds)) {
            return false;
        }
        for (int i = 0; i < upperBounds.length; ++i) {
            TypeBinding aBound = upperBounds[i];
            if ((aBound instanceof ReferenceBinding)) {
                if (this.superclass == null && aBound.isClass()) {
                    this.superclass = (ReferenceBinding)aBound;
                }
                else if (aBound.isInterface()) {
                    ++numReferenceInterfaces;
                }
            }
            else if (TypeBinding.equalsEquals(aBound.leafComponentType(), this)) {
                return false;
            }
        }
        this.superInterfaces = new ReferenceBinding[numReferenceInterfaces];
        int idx = 0;
        for (int j = 0; j < upperBounds.length; ++j) {
            TypeBinding aBound2 = upperBounds[j];
            if (aBound2.isInterface()) {
                this.superInterfaces[idx++] = (ReferenceBinding)aBound2;
            }
        }
        if (this.superclass == null) {
            this.superclass = javaLangObject;
        }
        return true;
    }
    
    @Override
    public void initializeBounds(Scope scope, ParameterizedTypeBinding capturedParameterizedType) {
    }
    
    @Override
    public TypeBinding clone(TypeBinding enclosingType) {
        return new CaptureBinding18(this);
    }
    
    @Override
    public MethodBinding[] getMethods(char[] selector) {
        if (this.upperBounds.length == 1 && this.upperBounds[0] instanceof ReferenceBinding) {
            return ((ReferenceBinding)this.upperBounds[0]).getMethods(selector);
        }
        return super.getMethods(selector);
    }
    
    @Override
    public TypeBinding erasure() {
        if (this.upperBounds != null && this.upperBounds.length > 1) {
            ReferenceBinding[] erasures = new ReferenceBinding[this.upperBounds.length];
            boolean multipleErasures = false;
            for (int i = 0; i < this.upperBounds.length; ++i) {
                erasures[i] = (ReferenceBinding)this.upperBounds[i].erasure();
                if (i > 0 && TypeBinding.notEquals(erasures[0], erasures[i])) {
                    multipleErasures = true;
                }
            }
            if (!multipleErasures) {
                return erasures[0];
            }
            return this.environment.createIntersectionType18(erasures);
        }
        else {
            return super.erasure();
        }
    }
    
    @Override
    public boolean isEquivalentTo(TypeBinding otherType) {
        if (TypeBinding.equalsEquals(this, otherType)) {
            return true;
        }
        if (otherType == null) {
            return false;
        }
        if (this.upperBounds != null) {
            for (int i = 0; i < this.upperBounds.length; ++i) {
                TypeBinding aBound = this.upperBounds[i];
                if (aBound != null && aBound.isArrayType()) {
                    if (!aBound.isCompatibleWith(otherType)) {
                        return false;
                    }
                }
                else {
                    switch (otherType.kind()) {
                        case 516:
                        case 8196: {
                            if (!((WildcardBinding)otherType).boundCheck(aBound)) {
                                return false;
                            }
                            break;
                        }
                    }
                }
            }
            return true;
        }
        return false;
    }
    
    @Override
    public boolean isCompatibleWith(TypeBinding otherType, Scope captureScope) {
        if (TypeBinding.equalsEquals(this, otherType)) {
            return true;
        }
        if (this.inRecursiveFunction) {
            return true;
        }
        this.inRecursiveFunction = true;
        try {
            if (this.upperBounds != null) {
                int length = this.upperBounds.length;
                int rightKind = otherType.kind();
                TypeBinding[] rightIntersectingTypes = null;
                if (rightKind == 8196 && otherType.boundKind() == 1) {
                    TypeBinding allRightBounds = ((WildcardBinding)otherType).allBounds();
                    if ((allRightBounds instanceof IntersectionTypeBinding18)) {
                        rightIntersectingTypes = ((IntersectionTypeBinding18)allRightBounds).intersectingTypes;
                    }
                }
                else if (rightKind == 32772) {
                    rightIntersectingTypes = ((IntersectionTypeBinding18)otherType).intersectingTypes;
                }
                if (rightIntersectingTypes != null) {
                    int numRequired = rightIntersectingTypes.length;
                    TypeBinding[] required = new TypeBinding[numRequired];
                    System.arraycopy(rightIntersectingTypes, 0, required, 0, numRequired);
                    for (int j = 0; j < length; ++j) {
                        TypeBinding provided = this.upperBounds[j];
                        for (int k = 0; k < required.length; ++k) {
                            if (required[k] != null && provided.isCompatibleWith(required[k], captureScope)) {
                                required[k] = null;
                                if (--numRequired == 0) {
                                    return true;
                                }
                                break;
                            }
                        }
                    }
                    return this.inRecursiveFunction = false;
                }
                int i = 0;
                while (i < length) {
                    if (this.upperBounds[i].isCompatibleWith(otherType, captureScope)) {
                        return true;
                    }
                    ++i;
                }
            }
            return this.inRecursiveFunction = false;
        }
        finally {
            this.inRecursiveFunction = false;
        }
    }
    
    @Override
    public TypeBinding findSuperTypeOriginatingFrom(TypeBinding otherType) {
        if (this.upperBounds != null && this.upperBounds.length > 1) {
            int i = 0;
            while (i < this.upperBounds.length) {
                TypeBinding candidate = this.upperBounds[i].findSuperTypeOriginatingFrom(otherType);
                if (candidate != null) {
                    return candidate;
                }
                ++i;
            }
        }
        return super.findSuperTypeOriginatingFrom(otherType);
    }
    
    @Override
    TypeBinding substituteInferenceVariable(InferenceVariable var, TypeBinding substituteType) {
        if (this.inRecursiveFunction) {
            return this;
        }
        this.inRecursiveFunction = true;
        try {
            boolean haveSubstitution = false;
            ReferenceBinding currentSuperclass = this.superclass;
            if (currentSuperclass != null) {
                currentSuperclass = (ReferenceBinding)currentSuperclass.substituteInferenceVariable(var, substituteType);
                haveSubstitution |= TypeBinding.notEquals(currentSuperclass, this.superclass);
            }
            ReferenceBinding[] currentSuperInterfaces = null;
            if (this.superInterfaces != null) {
                int length = this.superInterfaces.length;
                if (haveSubstitution) {
                    System.arraycopy(this.superInterfaces, 0, currentSuperInterfaces = new ReferenceBinding[length], 0, length);
                }
                for (int i = 0; i < length; ++i) {
                    ReferenceBinding currentSuperInterface = this.superInterfaces[i];
                    if (currentSuperInterface != null) {
                        currentSuperInterface = (ReferenceBinding)currentSuperInterface.substituteInferenceVariable(var, substituteType);
                        if (TypeBinding.notEquals(currentSuperInterface, this.superInterfaces[i])) {
                            if (currentSuperInterfaces == null) {
                                System.arraycopy(this.superInterfaces, 0, currentSuperInterfaces = new ReferenceBinding[length], 0, length);
                            }
                            currentSuperInterfaces[i] = currentSuperInterface;
                            haveSubstitution = true;
                        }
                    }
                }
            }
            TypeBinding[] currentUpperBounds = null;
            if (this.upperBounds != null) {
                int length2 = this.upperBounds.length;
                if (haveSubstitution) {
                    System.arraycopy(this.upperBounds, 0, currentUpperBounds = new TypeBinding[length2], 0, length2);
                }
                for (int j = 0; j < length2; ++j) {
                    TypeBinding currentBound = this.upperBounds[j];
                    if (currentBound != null) {
                        currentBound = currentBound.substituteInferenceVariable(var, substituteType);
                        if (TypeBinding.notEquals(currentBound, this.upperBounds[j])) {
                            if (currentUpperBounds == null) {
                                System.arraycopy(this.upperBounds, 0, currentUpperBounds = new TypeBinding[length2], 0, length2);
                            }
                            currentUpperBounds[j] = currentBound;
                            haveSubstitution = true;
                        }
                    }
                }
            }
            TypeBinding currentFirstBound = null;
            if (this.firstBound != null) {
                currentFirstBound = this.firstBound.substituteInferenceVariable(var, substituteType);
                haveSubstitution |= TypeBinding.notEquals(this.firstBound, currentFirstBound);
            }
            if (haveSubstitution) {
                final CaptureBinding18 newCapture = (CaptureBinding18)this.clone(this.enclosingType());
                newCapture.tagBits = this.tagBits;
                Substitution substitution = new Substitution() {
                    @Override
                    public TypeBinding substitute(TypeVariableBinding typeVariable) {
                        return (typeVariable == CaptureBinding18.this) ? newCapture : typeVariable;
                    }
                    
                    @Override
                    public boolean isRawSubstitution() {
                        return false;
                    }
                    
                    @Override
                    public LookupEnvironment environment() {
                        return CaptureBinding18.this.environment;
                    }
                };
                if (currentFirstBound != null) {
                    newCapture.firstBound = Scope.substitute(substitution, currentFirstBound);
                }
                newCapture.superclass = (ReferenceBinding)Scope.substitute(substitution, currentSuperclass);
                newCapture.superInterfaces = Scope.substitute(substitution, currentSuperInterfaces);
                newCapture.upperBounds = Scope.substitute(substitution, currentUpperBounds);
                return newCapture;
            }
            else {
                return this;
            }
        }
        finally {
            this.inRecursiveFunction = false;
        }
    }
    
    @Override
    public boolean isProperType(boolean admitCapture18) {
        if (!admitCapture18) {
            return false;
        }
        if (this.inRecursiveFunction) {
            return true;
        }
        this.inRecursiveFunction = true;
        try {
            if (this.lowerBound != null && !this.lowerBound.isProperType(admitCapture18)) {
                return this.inRecursiveFunction = false;
            }
            if (this.upperBounds != null) {
                int i = 0;
                while (i < this.upperBounds.length) {
                    if (!this.upperBounds[i].isProperType(admitCapture18)) {
                        return this.inRecursiveFunction = false;
                    }
                    ++i;
                }
            }
        }
        finally {
            this.inRecursiveFunction = false;
        }
        return true;
    }
    
    @Override
    public char[] genericTypeSignature() {
        if (this.genericTypeSignature == null) {
            try {
                char[] boundSignature;
                if (this.prototype.recursionLevel++ > 0 || this.firstBound == null) {
                    boundSignature = TypeConstants.WILDCARD_STAR;
                }
                else if (this.upperBounds != null) {
                    boundSignature = CharOperation.concat(TypeConstants.WILDCARD_PLUS, this.firstBound.genericTypeSignature());
                }
                else if (this.lowerBound != null) {
                    boundSignature = CharOperation.concat(TypeConstants.WILDCARD_MINUS, this.lowerBound.genericTypeSignature());
                }
                else {
                    boundSignature = TypeConstants.WILDCARD_STAR;
                }
                this.genericTypeSignature = CharOperation.concat(TypeConstants.WILDCARD_CAPTURE, boundSignature);
            }
            finally {
                CaptureBinding18 prototype = this.prototype;
                --prototype.recursionLevel;
            }
            CaptureBinding18 prototype2 = this.prototype;
            --prototype2.recursionLevel;
        }
        return this.genericTypeSignature;
    }
    
    @Override
    public char[] readableName() {
        if (this.lowerBound == null && this.firstBound != null) {
            if (this.prototype.recursionLevel < 2) {
                try {
                    CaptureBinding18 prototype = this.prototype;
                    ++prototype.recursionLevel;
                    if (this.upperBounds != null && this.upperBounds.length > 1) {
                        StringBuffer sb = new StringBuffer();
                        sb.append(this.upperBounds[0].readableName());
                        for (int i = 1; i < this.upperBounds.length; ++i) {
                            sb.append('&').append(this.upperBounds[i].readableName());
                        }
                        int len = sb.length();
                        char[] name = new char[len];
                        sb.getChars(0, len, name, 0);
                        CaptureBinding18 prototype3 = this.prototype;
                        --prototype3.recursionLevel;
                        return name;
                    }
                    char[] readableName = this.firstBound.readableName();
                    CaptureBinding18 prototype2 = this.prototype;
                    --prototype2.recursionLevel;
                    return readableName;
                }
                finally {
                    CaptureBinding18 prototype4 = this.prototype;
                    --prototype4.recursionLevel;
                }
            }
            return this.originalName;
        }
        else {
            return super.readableName();
        }
    }
    
    @Override
    public char[] shortReadableName() {
        if (this.lowerBound == null && this.firstBound != null) {
            if (this.prototype.recursionLevel < 2) {
                try {
                    CaptureBinding18 prototype = this.prototype;
                    ++prototype.recursionLevel;
                    if (this.upperBounds != null && this.upperBounds.length > 1) {
                        StringBuffer sb = new StringBuffer();
                        sb.append(this.upperBounds[0].shortReadableName());
                        for (int i = 1; i < this.upperBounds.length; ++i) {
                            sb.append('&').append(this.upperBounds[i].shortReadableName());
                        }
                        int len = sb.length();
                        char[] name = new char[len];
                        sb.getChars(0, len, name, 0);
                        CaptureBinding18 prototype3 = this.prototype;
                        --prototype3.recursionLevel;
                        return name;
                    }
                    char[] shortReadableName = this.firstBound.shortReadableName();
                    CaptureBinding18 prototype2 = this.prototype;
                    --prototype2.recursionLevel;
                    return shortReadableName;
                }
                finally {
                    CaptureBinding18 prototype4 = this.prototype;
                    --prototype4.recursionLevel;
                }
            }
            return this.originalName;
        }
        else {
            return super.shortReadableName();
        }
    }
    
    @Override
    public TypeBinding uncapture(Scope scope) {
        return this;
    }
    
    @Override
    public char[] computeUniqueKey(boolean isLeaf) {
        StringBuffer buffer = new StringBuffer();
        buffer.append(TypeConstants.CAPTURE18);
        buffer.append('{').append(this.end).append('#').append(this.captureID).append('}');
        buffer.append(';');
        int length = buffer.length();
        char[] uniqueKey = new char[length];
        buffer.getChars(0, length, uniqueKey, 0);
        return uniqueKey;
    }
}
